 package PeakParsing;

#
#   Author:     Gurmukh Sahota 07/15/2000
#   Copyright:  Gurmukh Sahota, 2000. All rights reserved.
#
#   Updated:  Gurmukh Sahota 04/15/2001
#             Added PAL, PRL and notes as well as the assignment parsing.
#

# Returns a reference to a Array     of  a Hash      of an Array
#                          Residue  :::  AtomTypes  :::  possible chemical shifts

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw( parsePeakfile outputPeakfile );
%EXPORT_TAGS = ( ALL => [@EXPORT_OK] );                                                   

# allows perl to fine local modules.
use FindBin;
use lib $FindBin::Bin;

use strict;


########################################## parsePeakfile #########################################################
# Input   : filename and tokens ie. dimensions used. Explained below                                             #
# Output  : A reference to a peakfile hash                                                                       #
#                $peakfile_hlist - reference to a hash                                                           #
#               $$peakfile_hlist{"filename"}    - peaklist's filename                                            #
#               $$peakfile_hlist{"dimlist"}      - the tokens of the peaklist or the dimension                   #
#                                                 names (HN, N15, CX)                                            #
#               $$peakfile_hlist{"internal_comments"}    - the comments in the peaklist, lines                   #
#                                                 starting with a #                                              #
#               $$peakfile_hlist{"end_comments"} - The filehistory created by some of the                        #
#                                                 AutoAssign Scripts                                             #
#               $$peakfile_hlist{"plist"}       - Hash list of index to the shifts and                           #
#                                                 intensity and workbooks                                        #
#               $peakfile_row = $$peakfile_hlist{"plist"}{indexnumber}                                           #
#                                                                                                                #
#                               $$peakfile_row{"shifts"}{token}               - shift value of the token         #
#                               $$peakfile_row{"intensity"}                   - intensity of the peak            #
#                               $$peakfile_row{"workbookname"}                - workbook name of the peak        #
#                               $$peakfile_row{"assign"}{"residue"}           - residue assignment of peak       #
#                               $$peakfile_row{"assign"}{"atomtype"}{"list"}  - atomType list for the assignment #
#                               $$peakfile_row{"PRL"}{"list"}                 - Possible Residue List(array)     #
#                               $$peakfile_row{"PAL"}{"list"}                 - Possible AtomType List(array)    #
#                                                                                                                #
# Purpose : To parse the input peakfile and return the described data structure                                  #
##################################################################################################################

sub parsePeakfile

  { #===== begin parsePeakfile =====#
    my $infile = shift @_;
    my @tokens = @_;

    my @input = ();
    my $count = 0;
    my $line;

    my $peakfile_hlist = {};
    $$peakfile_hlist{"internal_comments"}    = '';
    $$peakfile_hlist{"end_comments"} = '';

    # open the file for input
     local *IN;
     if ($infile =~ /^\-$/)
     {
	 *IN = *STDIN;
     }
     else
     {
	 open(IN, "<$infile");
     }
    # save the input parameters in the hash for future reference
    $$peakfile_hlist{"filename"} = $infile;
    $$peakfile_hlist{"dimlist"} = \@tokens;
    
    # while you do not hit the eof or catch a * do the following
    while ($line = <IN>)
      { 
	# save the comments and push the other lines on to an input array.
	if ($line =~ /^\*/)     {last;}
	elsif ($line =~ /^\#/)  {$$peakfile_hlist{"internal_comments"} .= $line;} 
	else                    {push (@input, $line);} 
      }
    
    # while there are extra lines, we will assume that they are file history lines
    while ($line = <IN>)
      {   
	$$peakfile_hlist{"end_comments"} .= $line; 
      } 

    # close the file
    close(IN);
    
    # while there are elements left in the input array
    while (@input > 0)
      {
	# pull off the next line
	$line = shift @input;

	# as long as the line is not empty or just spaces do
	if ($line !~ /^\s*$/)
	  {
	    # split the line on the spaces
	    my @chemical_shifts = split(/\s+/, $line);

	    # we assume that the first column is the index 
	    my $index = shift @chemical_shifts;

	    # the token columns are stored
	    for ($count=0; $count < @tokens; $count++)
	      {
		$$peakfile_hlist{"plist"}{$index}{"shifts"}{$tokens[$count]} = $chemical_shifts[$count];
	      }

	    # we assume that the next two columns will be intensity and the workbook name
	    $$peakfile_hlist{"plist"}{$index}{"intensity"} = $chemical_shifts[$count++];
	    $$peakfile_hlist{"plist"}{$index}{"workbookname"} = $chemical_shifts[$count++];
	    if ($$peakfile_hlist{"plist"}{$index}{"workbookname"} =~ /^(.*?)\.(\w\d+)\.?(.*?)\.(.*?)$/)
	    {
		$$peakfile_hlist{"plist"}{$index}{"assign"}{"residue"} = $2;
		@ {$$peakfile_hlist{"plist"}{$index}{"assign"}{"atomtype"}{"list"}} = (split /\_/, $3);
#		print "$index >> ", $$peakfile_hlist{"plist"}{$index}{"assign"}{"residue"}, "  :: (", join(", ", @ {$$peakfile_hlist{"plist"}{$index}{"assign"}{"atomtype"}{"list"}}), ")\n";
	    }
	    # We will assume that the note corresponds to a Possible Residue List
		while ( $$peakfile_hlist{"plist"}{$index}{"workbookname"} =~ /\{(\-?\d+),(.*?)\}/g)
		{
		    my $list_type = $1;
		    # PRL is a CSV list
		    @{$$peakfile_hlist{"plist"}{$index}{"PRL"}{$list_type}{"list"}} = split(/\,/, $2);
		    foreach my $elem (@{$$peakfile_hlist{"plist"}{$index}{"PRL"}{$list_type}{"list"}})
		    {
			$elem = uc($elem);
		    }
		    # Make sure something was in there ...
		    delete $$peakfile_hlist{"plist"}{$index}{"PRL"}{$list_type}{"list"} if (!(@{$$peakfile_hlist{"plist"}{$index}{"PRL"}{$list_type}{"list"}}));
#		    print $1, "\t", @{$$peakfile_hlist{"plist"}{$index}{"PRL"}{$list_type}{"list"}}, "\n";
		}

		if ($$peakfile_hlist{"plist"}{$index}{"workbookname"}  =~ /\[(.*?)\]/)
		{
		    # PAL is a CSV list
		    @{$$peakfile_hlist{"plist"}{$index}{"PAL"}{"list"}} = split(/\,/, $1);

		    foreach my $elem (@{$$peakfile_hlist{"plist"}{$index}{"PAL"}{"list"}})
		    {
			$elem = uc($elem);
		    }
		    # Make sure something was in there ...
		    delete $$peakfile_hlist{"plist"}{$index}{"PAL"}{"list"} if (!(@{$$peakfile_hlist{"plist"}{$index}{"PAL"}{"list"}}));
		}
	}
    }
    
    # returned the datastructure
    return $peakfile_hlist;

  } #===== end   parsePeakfile =====#			       


######################################### outputPeakfile #########################################################
# Input   : A reference to a peakfile hash                                                                       #
#                $peakfile_hlist - reference to a hash                                                           #
#               $$peakfile_hlist{"filename"}    - peaklist's filename                                            #
#               $$peakfile_hlist{"dimlist"}      - the tokens of the peaklist or the dimension                   #
#                                                 names (HN, N15, CX)                                            #
#               $$peakfile_hlist{"internal_comments"}    - the comments in the peaklist, lines                   #
#                                                 starting with a #                                              #
#               $$peakfile_hlist{"end_comments"} - The filehistory created by some of the                        #
#                                                 AutoAssign Scripts                                             #
#               $$peakfile_hlist{"plist"}       - Hash list of index to the shifts and                           #
#                                                 intensity and workbooks                                        #
#               $peakfile_row = $$peakfile_hlist{"plist"}{indexnumber}                                           #
#                                                                                                                #
#                               $$peakfile_row{"shifts"}{token}               - shift value of the token         #
#                               $$peakfile_row{"intensity"}                   - intensity of the peak            #
#                               $$peakfile_row{"workbookname"}                - workbook name of the peak        #
#                               $$peakfile_row{"assign"}{"residue"}           - residue assignment of peak       #
#                               $$peakfile_row{"assign"}{"atomtype"}{"list"}  - atomType list for the assignment #
#                               $$peakfile_row{"PRL"}{"list"}                 - Possible Residue List(array)     #
#                               $$peakfile_row{"PAL"}{"list"}                 - Possible AtomType List(array)    #
#                               $$peakfile_row{"notes"}                       - notes about a specific peak      #
#                                                                                                                #
# Output  : A peakfile                                                                                           # 
# Purpose : To output the described data structure                                                               #
##################################################################################################################

sub outputPeakfile

  { #===== begin outputPeakfile =====#
      my $peakfile_hlist = shift @_;
      my $filename = shift @_;
      
      $filename = $$peakfile_hlist{"filename"}      if ($filename eq '');
      local *OUTFILE;
      if (($filename eq "-") || ($filename eq ''))
	 { *OUTFILE = *STDOUT; }
      else
	 { open (OUTFILE, ">$filename") || die "unable to open $filename"; }
      

      print OUTFILE "#Index\t";
      for (my $x = 0; $x < @{$$peakfile_hlist{"dimlist"}}; $x++)
	  {
	      print OUTFILE $x+1 , "Dim\t";
	  }
      print OUTFILE "Intensity\tWorkbook\n";

      foreach my $index (sort { $a <=> $b } (keys %{$$peakfile_hlist{"plist"}}))
	  {
	      print OUTFILE "$index\t";
	      foreach my $token (@{$$peakfile_hlist{"dimlist"}})
		  {
		      print OUTFILE $$peakfile_hlist{"plist"}{$index}{"shifts"}{$token}, "\t";
		  }
	      print OUTFILE $$peakfile_hlist{"plist"}{$index}{"intensity"}, "\t", $$peakfile_hlist{"plist"}{$index}{"workbookname"};
	      if (exists $$peakfile_hlist{"plist"}{$index}{"PRL"})
	      {
		  print OUTFILE "!!{", join(",", $$peakfile_hlist{"plist"}{$index}{"PRL"}{"list"}), "\}!!";
	      }
	      
	      if (exists $$peakfile_hlist{"plist"}{$index}{"PAL"})
	      {
		  print OUTFILE "!![", join(",", $$peakfile_hlist{"plist"}{$index}{"PAL"}{"list"}), "\]!!";
	      }

	      print OUTFILE "\n";

	  }
      print OUTFILE "*\n";
      print OUTFILE $$peakfile_hlist{"internal_comments"}, "\n";
      print OUTFILE $$peakfile_hlist{"end_comments"}, "\n";

      return;
  }








